@@ -1,7 +1,7 @@
 package Complete::Bash;
 
-our $DATE = '2015-03-04'; # DATE
-our $VERSION = '0.17'; # VERSION
+our $DATE = '2015-04-01'; # DATE
+our $VERSION = '0.18'; # VERSION
 
 use 5.010001;
 use strict;
@@ -27,17 +27,86 @@ $SPEC{':package'} = {
     ],
 };
 
+sub _expand_tilde {
+    my ($user, $slash) = @_;
+    my @ent;
+    if (length $user) {
+        @ent = getpwnam($user);
+    } else {
+        @ent = getpwuid($>);
+        $user = $ent[0];
+    }
+    return $ent[7] . $slash if @ent;
+    "~$user$slash"; # return as-is when failed
+}
+
+sub _add_unquoted {
+    no warnings 'uninitialized';
+
+    my ($word, $is_cur_word, $after_ws) = @_;
+
+    #say "D:add_unquoted word=$word is_cur_word=$is_cur_word after_ws=$after_ws";
+
+    $word =~ s!^(~)(\w*)(/|\z) |  # 1) tilde  2) username  3) optional slash
+               \\(.)           |  # 4) escaped char
+               \$(\w+)            # 5) variable name
+              !
+                  $1 ? (not($after_ws) || $is_cur_word ? "$1$2$3" : _expand_tilde($2, $3)) :
+                      $4 ? $4 :
+                          ($is_cur_word ? "\$$5" : $ENV{$5})
+                              !egx;
+    $word;
+}
+
+sub _add_double_quoted {
+    no warnings 'uninitialized';
+
+    my ($word, $is_cur_word) = @_;
+
+    $word =~ s!\\(.)           |  # 1) escaped char
+               \$(\w+)            # 2) variable name
+              !
+                  $1 ? $1 :
+                      ($is_cur_word ? "\$$2" : $ENV{$2})
+                          !egx;
+    $word;
+}
+
+sub _add_single_quoted {
+    my $word = shift;
+    $word =~ s/\\(.)/$1/g;
+    $word;
+}
+
 $SPEC{parse_cmdline} = {
     v => 1.1,
     summary => 'Parse shell command-line for processing by completion routines',
     description => <<'_',
 
-This function basically converts COMP_LINE (str) and COMP_POINT (int) to become
-COMP_WORDS (array) and COMP_CWORD (int), like what bash supplies to shell
-functions. The differences with bash are: 1) quotes and backslashes are by
-default stripped, unless you specify `preserve_quotes`; 2) no word-breaking
-characters aside from whitespaces are used, unless you specify more
-word-breaking characters by setting `word_breaks`.
+This function basically converts COMP_LINE (str) and COMP_POINT (int) into
+something like (but not exactly the same as) COMP_WORDS (array) and COMP_CWORD
+(int) that bash supplies to shell functions.
+
+The differences with bash are (these differences are mostly for parsing
+convenience for programs that use this routine):
+
+1) quotes and backslashes are stripped (bash's COMP_WORDS contains all the
+quotes and backslashes);
+
+2) variables are substituted with their values from environment variables except
+for the current word (COMP_WORDS[COMP_CWORD]) (bash does not perform variable
+substitution for COMP_WORDS). However, note that special shell variables that
+are not environment variables like `$0`, `$_`, `$IFS` will not be replaced
+correctly because bash does not export those variables for us.
+
+3) tildes (~) are expanded with user's home directory except for the current
+word (bash does not perform tilde expansion for COMP_WORDS);
+
+4) no word-breaking characters aside from whitespaces and `=` are currently used
+(bash uses COMP_WORDBREAKS which by default also include `:`, `;`, and so on).
+This is done for convenience of parsing of Getopt::Long-based applications. More
+word-breaking characters might be used in the future, e.g. when we want to
+handle complex bash statements like pipes, redirection, etc.
 
 Caveats:
 
@@ -46,9 +115,8 @@ Caveats:
     % cmd --foo=bar
     % cmd --foo = bar
 
-Because they both expand to `['--foo', '=', 'bar']`, when `=` is used as a
-word-breaking character. But obviously `Getopt::Long` does not regard the two as
-equivalent.
+Because they both expand to `['--foo', '=', 'bar']`. But obviously
+`Getopt::Long` does not regard the two as equivalent.
 
 _
     args_as => 'array',
@@ -64,26 +132,6 @@ _
             schema => 'int*',
             pos => 1,
         },
-        word_breaks => {
-            summary => 'Extra characters to break word at',
-            description => <<'_',
-
-In addition to space and tab.
-
-Example: `=:`.
-
-Note that the characters won't break words if inside quotes or escaped.
-
-_
-            schema => 'str',
-            pos => 2,
-        },
-        preserve_quotes => {
-            summary => 'Whether to preserve quotes, like bash does',
-            schema => 'bool',
-            default => 0,
-            pos => 3,
-        },
     },
     result => {
         schema => ['array*', len=>2],
@@ -103,143 +151,84 @@ _
     },
     result_naked => 1,
     links => [
-        {
-            url => 'pm:Parse::CommandLine',
-            description => <<'_',
-
-The module `Parse::CommandLine` has a function called `parse_command_line()`
-which is similar, breaking a command-line string into words (in fact, currently
-`parse_cmdline()`'s implementation is stolen from this module). However,
-`parse_cmdline()` does not die on unclosed quotes and allows custom
-word-breaking characters.
-
-_
-        },
     ],
 };
 sub parse_cmdline {
-    my ($line, $point, $word_breaks, $preserve_quotes) = @_;
+    no warnings 'uninitialized';
+    my ($line, $point) = @_;
 
     $line  //= $ENV{COMP_LINE};
     $point //= $ENV{COMP_POINT} // 0;
-    $word_breaks //= '';
 
     die "$0: COMP_LINE not set, make sure this script is run under ".
         "bash completion (e.g. through complete -C)\n" unless defined $line;
 
-    my $pos = 0;
-    my $len = length($line);
-    # first word is ltrim-ed by bash
-    $line =~ s/\A(\s+)//gs and $pos += length($1);
-
     my @words;
-    my $buf;
     my $cword;
-    my $escaped;
-    my $inserted_empty_word;
-    my $double_quoted;
-    my $single_quoted;
+    my $pos = 0;
+    my $pos_min_ws = 0;
+    my $after_ws = 1;
+    my $chunk;
+    my $add_blank;
+    my $is_cur_word;
+    $line =~ s!(                                                 # 1) everything
+                  (")((?: \\\\|\\"|[^"])*)(?:"|\z)(\s*)       |  # 2) open "  3) content  4) space after
+                  (')((?: \\\\|\\'|[^'])*)(?:'|\z)(\s*)       |  # 5) open '  6) content  7) space after
+                  ((?: \\\\|\\"|\\'|\\=|\\\s|[^"'=\s])+)(\s*) |  # 8) unquoted word  9) space after
+                  = |
+                  \s+
+              )!
+                  $pos += length($1);
+                  #say "D:<$1> pos=$pos, point=$point, cword=$cword, after_ws=$after_ws";
 
-    my @chars = split //, $line;
-    $pos--;
-    for my $char (@chars) {
-        $pos++;
-        #say "D:pos=$pos, char=$char, \@words=[".join(", ", @words)."]";
-        if (!defined($cword) && $pos == $point) {
-            $cword = @words;
-            #say "D:setting cword to $cword";
-        }
+                  if ($2 || $5 || defined($8)) {
+                      # double-quoted/single-quoted/unquoted chunk
 
-        if ($escaped) {
-            $buf .= $preserve_quotes ? "\\$char" : $char;
-            $escaped = undef;
-            next;
-        }
+                      if (not(defined $cword)) {
+                          $pos_min_ws = $pos - length($2 ? $4 : $5 ? $7 : $9);
+                          #say "D:pos_min_ws=$pos_min_ws";
+                          if ($point <= $pos_min_ws) {
+                              $cword = @words - ($after_ws ? 0 : 1);
+                          } elsif ($point < $pos) {
+                              $cword = @words + 1 - ($after_ws ? 0 : 1);
+                              $add_blank = 1;
+                          }
+                      }
 
-        if ($char eq '\\') {
-            if ($single_quoted) {
-                $buf .= $char;
-            } else {
-                $escaped = 1;
-            }
-            next;
-        }
+                      if ($after_ws) {
+                          $is_cur_word = defined($cword) && $cword==@words;
+                      } else {
+                          $is_cur_word = defined($cword) && $cword==@words-1;
+                      }
+                      $chunk =
+                          $2 ? _add_double_quoted($3, $is_cur_word) :
+                              $5 ? _add_single_quoted($6) :
+                                  _add_unquoted($8, $is_cur_word, $after_ws);
+                      if ($after_ws) {
+                          push @words, $chunk;
+                      } else {
+                          $words[-1] .= $chunk;
+                      }
+                      if ($add_blank) {
+                          push @words, '';
+                          $add_blank = 0;
+                      }
+                      $after_ws = ($2 ? $4 : $5 ? $7 : $9) ? 1:0;
 
-        if ($char =~ /\s/) {
-            if ($single_quoted || $double_quoted) {
-                $buf .= $char;
-            } else {
-                if (defined $buf) {
-                    #say "D:pushing word <$buf>";
-                    push @words, $buf;
-                    undef $buf;
-                } elsif (!$inserted_empty_word &&
-                             $pos==$point && $chars[$pos-1] =~ /\s/ &&
-                                 $pos+1 < $len && $chars[$pos+1] =~ /\s/) {
-                    #say "D:insert empty word";
-                    push @words, '' unless $words[-1] eq '';
-                    $inserted_empty_word++;
-                }
-            }
-            next;
-        } else {
-            $inserted_empty_word = 0;
-        }
+                  } elsif ($1 eq '=') {
+                      # equal sign as word-breaking character
+                      push @words, '=';
+                      $after_ws = 1;
+                  } else {
+                      # whitespace
+                      $after_ws = 1;
+                  }
+    !egx;
 
-        if ($char eq '"') {
-            if ($single_quoted) {
-                $buf .= $char;
-                next;
-            }
-            $double_quoted = !$double_quoted;
-            if (!$double_quoted) {
-                $buf .= '"' if $preserve_quotes;
-            }
-            next;
-        }
+    $cword //= @words;
+    $words[$cword] //= '';
 
-        if ($char eq "'") {
-            if ($double_quoted) {
-                $buf .= $char;
-                next;
-            }
-            $single_quoted = !$single_quoted;
-            if (!$single_quoted) {
-                $buf .= "'" if $preserve_quotes;
-            }
-            next;
-        }
-
-        if (index($word_breaks, $char) >= 0) {
-            if ($escaped || $single_quoted || $double_quoted) {
-                $buf .= $single_quoted ? "'":'"' if !defined($buf) && $preserve_quotes;
-                $buf .= $char;
-                next;
-            }
-            push @words, $buf if defined $buf;
-            push @words, $char;
-            undef $buf;
-            next;
-        }
-
-        $buf .= $single_quoted ? "'" : $double_quoted ? '"' : '' if !defined($buf) && $preserve_quotes;
-        $buf .= $char;
-    }
-
-    if (defined $buf) {
-        #say "D:pushing last word <$buf>";
-        push @words, $buf;
-        $cword //= @words-1;
-    } else {
-        if (!@words || $words[-1] ne '') {
-            $cword //= @words;
-            $words[$cword] //= '';
-        } else {
-            $cword //= @words-1;
-        }
-    }
-
-    return [\@words, $cword];
+    [\@words, $cword];
 }
 
 $SPEC{parse_options} = {
@@ -541,7 +530,7 @@ Complete::Bash - Completion module for bash shell
 
 =head1 VERSION
 
-This document describes version 0.17 of Complete::Bash (from Perl distribution Complete-Bash), released on 2015-03-04.
+This document describes version 0.18 of Complete::Bash (from Perl distribution Complete-Bash), released on 2015-04-01.
 
 =head1 DESCRIPTION
 
@@ -661,16 +650,34 @@ Either an array or hash. See function description for more details.
 Return value: Formatted string (or array, if `as` is set to `array`) (str|array)
 
 
-=head2 parse_cmdline($cmdline, $point, $word_breaks, $preserve_quotes) -> array
+=head2 parse_cmdline($cmdline, $point) -> array
 
 Parse shell command-line for processing by completion routines.
 
-This function basically converts COMP_LINE (str) and COMP_POINT (int) to become
-COMP_WORDS (array) and COMP_CWORD (int), like what bash supplies to shell
-functions. The differences with bash are: 1) quotes and backslashes are by
-default stripped, unless you specify C<preserve_quotes>; 2) no word-breaking
-characters aside from whitespaces are used, unless you specify more
-word-breaking characters by setting C<word_breaks>.
+This function basically converts COMP_LINE (str) and COMP_POINT (int) into
+something like (but not exactly the same as) COMP_WORDS (array) and COMP_CWORD
+(int) that bash supplies to shell functions.
+
+The differences with bash are (these differences are mostly for parsing
+convenience for programs that use this routine):
+
+1) quotes and backslashes are stripped (bash's COMP_WORDS contains all the
+quotes and backslashes);
+
+2) variables are substituted with their values from environment variables except
+for the current word (COMP_WORDS[COMP_CWORD]) (bash does not perform variable
+substitution for COMP_WORDS). However, note that special shell variables that
+are not environment variables like C<$0>, C<$_>, C<$IFS> will not be replaced
+correctly because bash does not export those variables for us.
+
+3) tildes (~) are expanded with user's home directory except for the current
+word (bash does not perform tilde expansion for COMP_WORDS);
+
+4) no word-breaking characters aside from whitespaces and C<=> are currently used
+(bash uses COMP_WORDBREAKS which by default also include C<:>, C<;>, and so on).
+This is done for convenience of parsing of Getopt::Long-based applications. More
+word-breaking characters might be used in the future, e.g. when we want to
+handle complex bash statements like pipes, redirection, etc.
 
 Caveats:
 
@@ -683,9 +690,8 @@ Caveats:
 
 =back
 
-Because they both expand to C<['--foo', '=', 'bar']>, when C<=> is used as a
-word-breaking character. But obviously C<Getopt::Long> does not regard the two as
-equivalent.
+Because they both expand to C<['--foo', '=', 'bar']>. But obviously
+C<Getopt::Long> does not regard the two as equivalent.
 
 Arguments ('*' denotes required arguments):
 
@@ -699,20 +705,6 @@ Command-line, defaults to COMP_LINE environment.
 
 Point/position to complete in command-line, defaults to COMP_POINT.
 
-=item * B<preserve_quotes> => I<bool> (default: 0)
-
-Whether to preserve quotes, like bash does.
-
-=item * B<word_breaks> => I<str>
-
-Extra characters to break word at.
-
-In addition to space and tab.
-
-Example: C<=:>.
-
-Note that the characters won't break words if inside quotes or escaped.
-
 =back
 
 Return value:  (array)
@@ -727,20 +719,6 @@ Note that COMP_LINE includes the command name. If you want the command-line
 arguments only (like in C<@ARGV>), you need to strip the first element from
 C<$words> and reduce C<$cword> by 1.
 
-See also:
-
-=over
-
-* L<Parse::CommandLine>
-
-The module C<Parse::CommandLine> has a function called C<parse_command_line()>
-which is similar, breaking a command-line string into words (in fact, currently
-C<parse_cmdline()>'s implementation is stolen from this module). However,
-C<parse_cmdline()> does not die on unclosed quotes and allows custom
-word-breaking characters.
-
-=back
-
 
 =head2 parse_options(%args) -> [status, msg, result, meta]
 
@@ -828,7 +806,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/Complete-B
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/perlancar/perl-Complete-Bash>.
+Source repository is at L<https://github.com/sharyanto/perl-Complete-Bash>.
 
 =head1 BUGS
 
